import ftplib
import json
import io
import random
import ssl
import threading
import time
import zipfile

import paho.mqtt.client as mqtt

import base_sender


class P1P_FTP_TLS(ftplib.FTP_TLS):
    """FTP_TLS subclass that automatically wraps sockets in SSL to support FTPS."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._sock = None

    @property
    def sock(self):
        """Return the socket."""
        return self._sock

    @sock.setter
    def sock(self, value):
        """When modifying the socket, ensure that it is ssl wrapped."""
        if value is not None and not isinstance(value, ssl.SSLSocket):
            value = self.context.wrap_socket(value)
        self._sock = value


class X1_FTP_TLS(P1P_FTP_TLS):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    """Explicit FTPS, with shared TLS session"""
    def ntransfercmd(self, cmd, rest=None):
        conn, size = ftplib.FTP.ntransfercmd(self, cmd, rest)
        if self._prot_p:
            conn = self.context.wrap_socket(conn,
                                            server_hostname=self.host,
                                            session=self.sock.session)  # this is the fix
        return conn, size


class Sender(base_sender.BaseSender):
    MQTT_PORT = 8883
    FTP_PORT = 990
    USERNAME = "bblp"
    LOOP_SLEEP_SEC = 5
    FTP_TIMEOUT = 15
    PROJECT_PRINT_PATH = "Metadata/plate_1.gcode"

    CONNECT_RC_DESC = {
        0: "Connection successful",
        1: "Connection refused - incorrect protocol version",
        2: "Connection refused - invalid client identifier",
        3: "Connection refused - server unavailable",
        4: "Connection refused - bad username or password",
        5: "Connection refused - not authorised"
    }

    DEFAULT_SETTINGS = {
        "use_ams" : True,
        "timelapse" : False,
        "bed_leveling" : True,
        "flow_cali" : False,
        "vibration_cali" : False,
        "layer_inspect" : True
    }

    def __init__(self, parent, usb_info, profile={}):
        super().__init__(parent, usb_info, profile)
        self.serial_number = usb_info.get("SNR", "")
        self.host = usb_info.get("IP", "")
        self.run_once = usb_info.get('RUNONCE')
        self.password = usb_info.get('PASS')
        self.client = mqtt.Client()
        self.client.on_connect = self.on_connect
        self.client.on_message = self.on_message
        self.last_message = ""
        self.user_id = '3DPrinterOS_Cloud_Client' + str(random.randint(1000, 9999))
        self.logger.info(f'UserID for BambuLab requests for {str(self.host)}: {self.user_id}')
        self.current_sequance_id = 0
        self.printer_has_sdcard = False
        # Set the TLS/SSL options for the client
        self.client.tls_set(tls_version=ssl.PROTOCOL_TLS, cert_reqs=ssl.CERT_NONE)
        self.client.tls_insecure_set(True)
        self.client.username_pw_set(self.USERNAME, self.password)
        self.status_thread = threading.Thread(target=self.status_loop, daemon=True)
        self.publish_lock = threading.RLock()
        self.material_list = []
        self.has_ams = False
        self.mqtt_device_prefix = "device/%s/" % self.serial_number
        self.settings.update(self.DEFAULT_SETTINGS)
        self.load_printer_settings()
        if not self.run_once:
            self.connect_loop()

    def on_connect(self, client, userdata, flags, rc):
        self.logger.info(f"Bambulab: Connected to {self.host} with result code {rc}")
        if rc != 0:
            if rc in self.CONNECT_RC_DESC:
                desc = self.CONNECT_RC_DESC[rc]
            else:
                desc = "Return code " + str(rc)
            message = f"Error connecting Bambu Lab Printer on {self.host}. Description: {desc}"
            self.logger.error(message)
            self.client.disconnect()
            if self.parent:
                self.parent.register_error(1001, message)
            return
        client.subscribe(self.mqtt_device_prefix + 'report')
        self.send_get_version()
        self.status_thread.start()

    def on_disconnect(self):
        self.stop_flag = True
        self.logger.info("BambuLab: Disconnected from " + str(self.host))

    def on_message(self, client, userdata, message):
        response = json.loads(message.payload.decode())
        if response:
            if response.get("print"):
                self.logger.debug(f"Received from {self.host}. Desc: {response}")
                # TODO: 3mf parsing error code 83902467
                if response.get("print").get("command") == "push_status":
                    self.operational_flag = True
                    # parse AMS materials names
                    ams = response.get("print").get("ams", {}).get("ams", [])
                    self.has_ams = len(ams) > 0
                    if self.has_ams:
                        self.material_list = []
                        for item in ams:
                            ams_trays = item.get("tray", [])
                            for tray in ams_trays:
                                self.material_list.append(tray.get("tray_type"))
                    self.temps[0] = response.get("print").get("bed_temper", self.temps[0])
                    self.temps[1] = response.get("print").get("nozzle_temper", self.temps[1])
                    self.target_temps[0] = response.get("print").get("bed_target_temper", self.target_temps[0])
                    self.target_temps[1] = response.get("print").get("nozzle_target_temper", self.target_temps[1])
                    self.printer_has_sdcard = response.get("print").get("sdcard", self.printer_has_sdcard)

                    if "gcode_state" in response.get("print", {}): # need to check if key exist first for P1P model
                        gcode_state = response.get("print").get("gcode_state")
                        self.printing_flag = gcode_state == "RUNNING"
                        self.pause_flag = gcode_state == "PAUSE"
                        if self.is_printing() or self.is_paused():
                            self.estimated_time = response.get("print").get("mc_remaining_time")
                            self.percent = response.get("print").get("mc_percent")
                            self.current_line_number = response.get("print").get("layer_num")
                        else:
                            self.percent = 0
                            self.current_line_number = 0
            else:
                self.logger.debug(f"Can't parse response from {self.host}. Desc: {response}")

    def get_material_names(self):
        return self.material_list

    def status_loop(self):
        while not self.stop_flag and self.client.is_connected():
            self.send_pushall()
            time.sleep(self.LOOP_SLEEP_SEC)
        self.stop_flag = True
        self.operational_flag = False
        self.client.disconnect()
        self.logger.info(f"Status loop exit for BambuLab on {self.host}")

    def connect_loop(self):
        self.logger.info(f"Connecting Bambu Lab printer on {self.host}")
        try:
            self.client.connect(self.host, self.MQTT_PORT)
            self.client.loop_start()
        except OSError as e:
            msg = f"Error connecting Bambu Lab printer on {self.host}. Description: {e}"
            self.logger.error(msg)
            if self.parent:
                self.parent.register_error(1002, msg)

    def publish(self, msg):
        with self.publish_lock:
            result = self.client.publish(self.mqtt_device_prefix + "request", json.dumps(msg))
            status = result[0]
            if status == 0:
                self.logger.debug(f"Sent {msg} to topic {self.mqtt_device_prefix} request")
                return True
            msg = f"Failed to send message {msg} to topic {self.mqtt_device_prefix} request on {self.host}."
            self.logger.warning(msg)
            if self.parent:
                self.parent.register_error(1003, msg)
            return False

    def send_print_command(self, command, param=None):
        self.current_sequance_id += 1
        request = {"print": {"sequence_id": str(self.current_sequance_id), "command": command}, "user_id": self.user_id}
        if param:
            request["print"]["param"] = param
        return self.publish(request)

    def send_pushall(self):
        self.current_sequance_id += 1
        request = {"pushing": {"sequence_id": str(self.current_sequance_id), "command": "pushall"}, "user_id": self.user_id}
        return self.publish(request)

    def send_get_version(self):
        self.current_sequance_id += 1
        request = {"info": {"sequence_id": str(self.current_sequance_id), "command": "get_version"}, "user_id": self.user_id}
        return self.publish(request)

    def pause(self):
        return self.send_print_command("pause")

    def unpause(self):
        return self.send_print_command("resume")

    def cancel(self):
        return self.send_print_command("stop")

    def create_project(self, gcode):
        zip_data = io.BytesIO()
        with zipfile.ZipFile(zip_data, 'w', zipfile.ZIP_DEFLATED) as zipf:
            zipf.writestr('Metadata/', '')
            zipf.writestr(self.PROJECT_PRINT_PATH, gcode.read())
        zip_data.seek(0)
        return zip_data

    def create_ftp_tls(self):
        if self.usb_info.get("PID") == "0P1P":
            return P1P_FTP_TLS()
        else:
            return X1_FTP_TLS() #TODO: text for X1

    def check_ftp_file(self, local_file, ftp_file):
        try:
            ftp_session = self.create_ftp_tls()
            ftp_session.encoding = "utf-8"
            ftp_session.timeout = self.FTP_TIMEOUT
            ftp_session.connect(host=self.host, port=self.FTP_PORT, timeout=self.FTP_TIMEOUT)
            ftp_session.login(self.USERNAME, self.password)
            ftp_session.prot_p()
            ftp_session.cwd('/')
            file_exists = False

            def set_file_exists(line):
                nonlocal file_exists
                if ftp_file in line:
                    file_exists = True

            ftp_session.dir(set_file_exists)
            if not file_exists:
                self.logger.info(f"File {ftp_file} not found on FTP server")
                ftp_session.quit()
                return False
            ftp_size = ftp_session.size(ftp_file)
            local_size = len(local_file.getvalue())
            if ftp_size != local_size:
                self.logger.warning(f"File {ftp_file} has different size on FTP server")
                ftp_session.quit()
                return False
            self.logger.warning(f"File {ftp_file} found on FTP server with the same size")
            ftp_session.quit()
            return True
        except ftplib.all_errors as err:
            msg = f"Check file on {self.host} failed. Description: {str(err)}"
            self.logger.warning(msg)
            if self.parent:
                self.parent.register_error(1006, msg)
        return False

    def upload(self, project_file, file_name):
        if self.printer_has_sdcard:
            self.logger.info(f"Start uploading {file_name} to {self.host}")
            ftp_session = self.create_ftp_tls()
            try:
                ftp_session.encoding = "utf-8"
                ftp_session.timeout = self.FTP_TIMEOUT
                ftp_session.connect(host=self.host, port=self.FTP_PORT, timeout=self.FTP_TIMEOUT)
                ftp_session.login(self.USERNAME, self.password)
                ftp_session.prot_p()
                ftp_session.storbinary(f"STOR {file_name}", project_file)
                self.logger.info(f"Successful uploaded {file_name} on {self.host}")
                ftp_session.quit()
                return True
            except ftplib.all_errors as err:
                if self.check_ftp_file(project_file, file_name):
                    return True
                msg = f"Upload to {self.host} failed. Description: {str(err)}"
                self.logger.warning(msg)
                if self.parent:
                    self.parent.register_error(1005, msg)
        else:
            msg = f"Can't upload file on {self.host}. Description: no sdcard"
            self.logger.warning(msg)
            if self.parent:
                self.parent.register_error(1004, msg)
        return False

    # NOTE: command from Bambulab Studio API Source. Not working with files for print (Tested on X1C Carbon)
    def run_gcode_file(self, sdcard_filepath):
        return self.send_print_command("gcode_file", sdcard_filepath)

    def unbuffered_gcodes(self, gcodes):
        self.logger.info(f"Enter unbuffered_gcodes: {gcodes}")
        try:
            gcodes = self.preprocess_gcodes(gcodes)
        except Exception as e:
            self.logger.exception(e)
            return False
        success = True
        for command in gcodes:
            request = {"print": {"command": "gcode_line","param": command},"user_id": self.user_id}
            success &= self.publish(request)
        return success

    def gcodes(self, gcodes_or_file):
        if isinstance(gcodes_or_file, io.IOBase):
            project_file = self.create_project(gcodes_or_file)
            project_filename = self.get_filename() + '.3mf'
            if self.upload(project_file, project_filename):
                self.current_sequance_id += 1
                request = {
                    "print": {
                        "sequence_id": self.current_sequance_id,
                        "command": "project_file",
                        "param": self.PROJECT_PRINT_PATH,
                        "subtask_name": self.get_filename(),  
                        "url": f"ftp://{project_filename}",
                        "use_ams": self.has_ams and self.settings['use_ams']
                    }
                }
                if self.settings != self.DEFAULT_SETTINGS:
                    request["print"].update(self.settings)
                return self.publish(request)
        else:
            return self.unbuffered_gcodes(gcodes_or_file)
